/*
 * common_app.c
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler and device-autorotator are free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 * 
 * touchscreen-autodisabler and device-autorotator are distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */
/*
 * Provides feature for daemon like application.
 *
 * touchscreen_autodisabler (7) and device_autorotator (7) will share
 * this file. (This means, they will contain this as utility functionality.)
 */

#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <pwd.h>
#include <signal.h>

#include <errno.h>

#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <stdarg.h>

#include "common_util.h"
#include "common_app.h"

#include <libintl.h>
#define _(s)	gettext(s)

/* callback type declaration **************************************************/

/*
 * OptionFunc:
 * @option_struct_member: (out):	pointer to value to set.
 * @type_data:						data for type.
 * @option_arg:						option argument to process.
 * 
 * This is common prototype for command line option process function.
 * For each type, a process function will be used.
 * 
 * Processing function should process option_arg, set option_struct_member.
 * On error it should return %false.
 * 
 * Returns: if option process was succesful.
 */
typedef bool	(*OptionFunc)   (void*			option_struct_member,
	                             const void*	type_data,
	                             const char*	option_arg);

/* function declarations - application core support - internal ****************/

static void		common_app_at_exit_cleanup			(void);

static bool		common_app_pid_file_claim_file		(CommonPidFile* pid_file,
				                                     const char*	write_str);
static bool		common_app_pid_file_read_file		(CommonPidFile* pid_file);

static void		common_app_pid_file_add_cleanup		(CommonPidFile* pid_file);
static void		common_app_pid_file_remove_cleanup	(CommonPidFile* pid_file);
static void		common_app_pid_file_cleanup_exit	(void);

/* function declarations - option processing - internal ***********************/

static bool		common_option_process_void			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

static bool		common_option_process_bool			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

static bool		common_option_process_char			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

static bool		common_option_process_int			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

static bool		common_option_process_enum			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

static bool		common_option_process_str			(void*			option_ptr,
							                         const void*	type_data,
							                         const char*	opt_arg);

/* constant declarations - option processing - internal ***********************/

static const OptionFunc option_funcs[COMMON_TYPE_COUNTS] = {
	common_option_process_void,
	common_option_process_bool,
	common_option_process_char,
	common_option_process_int,
	common_option_process_enum,
	common_option_process_str
};
/* variables - application core support ***************************************/

static CommonClosure*   at_exit_closure				= NULL;

static int				at_exit_count				= 0;
static int				at_exit_alloc				= 0;

static CommonPidFile**	pid_file_cleanup			= NULL;
static int				pid_file_cleanup_count		= 0;
static int				pid_file_cleanup_alloc		= 0;

static bool				cleanup_mode				= false;


/* function body - application core support ***********************************/

/*
 * common_app_at_exit:
 * @func:	   function to be invoked at exit()
 * @user_data: user data to be supplied to @func
 * @free_func: function to be invoked to cleanup @user_data.
 *
 * Register @func to be invoked at invocation of common_app_at_exit().
 *
 * As common_app_at_exit() is registered with atexit(), @func will be invoked
 * at exit().
 *
 * user_data should remain vaild until invocation of @func. (generally at exit())
 *
 * This has two point of advantage over atexit().
 * 1. Unlimited number of registered functions.
 *  atexit() has limit of registered functions. but common_app_at_exit() 
 *  registers function at dynamic allocated array, so it can hold unlimited
 *  number of functions as memory allows.
 * 
 * 2. User data can be sent.
 *  on atexit(), caller cannot send user data. but this can register @func with
 *  @user_data. So, exit processing logic doesn't introduce global/static
 *  variables. Evenmore, if you need separate cleanup code for @user_data,
 *  you can put it as @free_func.
 */
void
common_app_at_exit	(ExitFunc			func,
					 void*				user_data,
	                 CommonFreeFunc		free_func)
{
	if (at_exit_alloc == 0) {
		at_exit_alloc = 4;
		at_exit_closure = (CommonClosure*) malloc (at_exit_alloc * sizeof (CommonClosure));
		atexit (common_app_at_exit_cleanup);
	}

	else if (at_exit_alloc == at_exit_count) {
		at_exit_alloc *= 2;
		at_exit_closure = (CommonClosure*) realloc (at_exit_closure,
		                                            at_exit_alloc * sizeof (CommonClosure));
	}

	common_closure_set (& (at_exit_closure[at_exit_count++]),
	                    func, user_data, free_func);
}


/*
 * common_app_register_sig:
 * @func:   function to be called at occurance of signal.
 * @...:	signal to register @func.
 * 
 * Registers @func for @... signals.
 */
bool
common_app_register_sig (void (*func)(int),
                         ...)   //expecting int, ... , 0
{
	struct sigaction	sigact;
	
	va_list				va_args;
	int					va_elem;
	
	sigact.sa_handler = func;

	sigemptyset (&sigact.sa_mask);
	sigact.sa_flags = 0;
	
	va_start (va_args, func);
	while ((va_elem = va_arg(va_args, int)) != 0)
		if (sigaction (va_elem, &sigact, NULL) != 0) {
			return false;
			va_end (va_args);
		}
	va_end (va_args);
	
	return true;
}

/*
 * common_app_home_path:
 * 
 * Returns: (transfer full): user's home folder path.
 */
char*
common_app_home_path (void)
{
	struct passwd*	pwd;
	char*			result;

	result = getenv("HOME");

	if (result == NULL) {
		pwd = getpwuid (getuid ());
		result = pwd->pw_dir;
	}
	return result;
}

/*
 * common_app_pid_file:
 * @lock_file_path:	file path for pid file.
 * @write_str:		additional string to write pid file.
 * 
 * Try open up pid file at @lock_file_path, and write current pid.
 * You can check #CommonPidFile.holder to see if pid file is locked or not.
 * 
 * Returns: CommonPidFile struct.
 */
CommonPidFile*
common_app_pid_file		(const char*	lock_file_path,
		                 const char*	write_str,
		                 bool			replace)
{
	CommonPidFile*  result;

	bool			die;

	
	result = (CommonPidFile*) malloc (sizeof (CommonPidFile));

	result->file_path = common_strdup (lock_file_path);
	result->fd = open (lock_file_path, O_RDWR | O_CREAT, 0600);
	result->holder = 0;
	result->rest_str = NULL;

	// Problem in openning/creating file.
	if (result->fd < 0)
		die = true;

	// Lock file acquired
	else if (common_app_pid_file_claim_file (result, write_str))
		die = false;

	// File has locked by other process.
	else if (errno == EAGAIN) {
		if (common_app_pid_file_read_file (result)) {
			if (replace) {
				close (result->fd);
				common_app_replace (result->holder);
				result->fd = open (lock_file_path, O_RDWR | O_CREAT, 0600);

				common_app_pid_file_claim_file (result, write_str);
			}
			die = false;
		}
		else die = true;
	}
	
	// Error!
	else die = true;


	// Return results 
	if (die) {
		common_app_pid_file_free (result);
		return NULL;
	}
	else {
		common_app_pid_file_add_cleanup (result);
		return result;
	}
}


/*
 * common_app_pid_file_free:
 * @file: representative struct for pid file.
 * 
 * Close file description and deletes pid file.
 */
void
common_app_pid_file_free (CommonPidFile*	file)
{
	close (file->fd);

	if (file->holder == getpid()) unlink (file->file_path);
	free (file->file_path);

	if (file->rest_str != NULL) free (file->rest_str);

	if (!cleanup_mode) common_app_pid_file_remove_cleanup (file);
	free (file);
}

/*
 * common_app_daemonize:
 * @child:  pointer to retrieve pid of child process.
 * 
 * Spawns a daemonized child process. This will not exit parent process, but
 * returns %COMMON_DAEMON_RESULT_PARENT so parent can tide up before exiting.
 * 
 * Returns: %COMMON_DAEMON_RESULT_OK to child,
 *			%COMMON_DAEMON_RESULT_PARENT to parent,
 *			%COMMON_DAEMON_RESULT_FAIL if fail.
 */
CommonDaemonResult
common_app_daemonize (pid_t*   child)
{
	pid_t   pid;
	pid_t   sid;

	if (getppid() == 1) return COMMON_DAEMON_RESULT_OK;

	pid = fork ();

	if (pid < 0) return COMMON_DAEMON_RESULT_FAIL;
	else if (pid > 0) {
		*child = pid;
		return COMMON_DAEMON_RESULT_PARENT;
	}

	umask (0);
	sid = setsid();

	if (sid < 0) return COMMON_DAEMON_RESULT_FAIL;
	if ((chdir("/")) < 0) return COMMON_DAEMON_RESULT_FAIL;

	freopen ("/dev/null", "r", stdin);
	freopen ("/dev/null", "w", stdout);
	freopen ("/dev/null", "w", stderr);

	return COMMON_DAEMON_RESULT_OK;
}

/*
 * common_app_replace:
 * @instance:   instance to replace.
 * 
 * exit @instance.
 */
void
common_app_replace (pid_t instance)
{
	long	timeout;
	long	step;
	
	timeout = 5000000;
	step = 1000;

	kill (instance, SIGINT);
	
	while (timeout > 0) {
		usleep (step);
		if (kill (instance, 0)) break;
		else timeout -= step;
	}
	if (timeout <= 0) {
		
		kill (instance, SIGKILL);
		usleep (step);
	}
}

/* function body - option processing ******************************************/

/*
 * common_option_generate_short:
 * @option_longopt: struct for getopt_long()
 *
 * generate short option description from @option_longopt.
 * generated option requires GNU Extension for optional argument of short opts.
 * 
 * Returns: short option string that can be used for getopt()
 */
char*
common_option_generate_short(const struct option*		option_longopt)
{
	char*   result;
	char*   result_p;
	size_t  length;

	const struct option*  option_p;

	length = 0;
	for (option_p = option_longopt; option_p->name != NULL; option_p++) {
		if (isalpha (option_p->val)) {
			length += (option_p->has_arg == no_argument) ? 0 :
					  (option_p->has_arg == required_argument) ? 1 : 2;
		}
	}

	result = (char*) malloc (length * sizeof (char));
	for (result_p = result, option_p = option_longopt;  option_p->name != NULL; option_p++) {
		if (isalpha (option_p->val)) {
			*(result_p++) = (char)option_p->val;
			if (option_p->has_arg != no_argument) {
				*(result_p++) = ':';
				if (option_p->has_arg == optional_argument)
					*(result_p++) = ':';
			}
		}
	}
	*result_p = '\0';
	return result;
}

/*
 * common_option_process:
 * @option_struct: (out):   struct to store options.
 * @option_longopt:			struct for getopt_long()
 * @option_item:			option items.
 * @argc:					argument count from main()
 * @argv: (array length=argc):  arguments from main()
 *
 * Processes options and put result in @option_struct.
 * 
 * Returns: %true on success.
 */
bool
common_option_process	(void*						option_struct,
                         const struct option*		option_longopt,
                         const CommonOptionItem*	option_item,
                         int						argc,
                         char**						argv)
{
	char*   option_short;

	int		option_index;

	int		option_res;

	void*   option_struct_member;
	const CommonOptionItem*   option_curr_item;

	((CommonOptionResultBase*)option_struct)->command_line =
		common_str_join (argv + 1, argc - 1, ' ');

	option_short = common_option_generate_short (option_longopt);
	// Parse options.
	do {
		option_index = -1;
		option_res = getopt_long (argc, argv, option_short, option_longopt, &option_index);

		// Determine opt_res.
		if (option_res == '?') option_res = optopt;
		else if (option_res == -1) break;

		// Gets opt_index if option was given in short form.
		if (option_index == -1)
			for (option_index = 0;
			     option_longopt[option_index].name != NULL;
			     option_index++)
				if (option_longopt[option_index].val == option_res) break;
		
		if (option_longopt[option_index].name == NULL) {
			free (option_short);
			return false;
		}

		option_curr_item = option_item + option_index;
		option_struct_member = option_struct + option_curr_item->offset;

		if (! option_funcs [option_curr_item->type] (option_struct_member,
		                                             option_curr_item->type_data,
				                                     optarg)) {
			free(option_short);
			return false;
		}
	}
	while (option_res != -1);

	free (option_short);
	return true;
}

/*
 * common_option_help:
 * @option_longopt: structure for getopt_long()
 * @option_item:	option item.
 * 
 * Prints option help using @option_longopt and @option_item.
 */
void
common_option_help  (const struct option*		option_longopt,
                     const CommonOptionItem*  option_item)
{
	int i;
	int j;

	char opt_buf[31];

	CommonOptionEnumData*   edata;

	for (i = 0; option_longopt[i].name != NULL; i++) {
		if (isalpha (option_longopt[i].val))
			sprintf (opt_buf,
			         "--%-22s, -%c",
			         option_longopt[i].name,
			         option_longopt[i].val);
		else
			sprintf (opt_buf,
			         "--%s",
			         option_longopt[i].name);

		printf ("%-30s: %s\n", opt_buf, gettext(option_item[i].description));

		if (option_item[i].type == COMMON_TYPE_ENUM) {
			edata = (CommonOptionEnumData*) option_item[i].type_data;
			printf (_("%32sPossible value: "), "");
			fputs (edata->str_list[0], stdout);
			for (j = 1; j < edata->str_list_count; j++)
				printf (", %s", edata->str_list[j]);
		}
	}
}


/* function body - application core support - internal ************************/

static void
common_app_at_exit_cleanup (void)
{
	int i;

	for (i = at_exit_count - 1; 0 <= i; i--)
		COMMON_CLOSURE_INV_VOID (void, (& at_exit_closure[i]));

	free (at_exit_closure);
}


static bool
common_app_pid_file_claim_file (CommonPidFile*  pid_file,
                                const char*		write_str)
{
	char*			buf;
	int				buf_len;
	
	struct flock	fl;
	
	int pid_inst;
	
	fl.l_start = 0;
	fl.l_len = 0;
	fl.l_type = F_WRLCK;
	fl.l_whence = SEEK_SET;

	if (fcntl (pid_file->fd, F_SETLK, &fl) < 0) return false;
	
	
	pid_inst = (int) getpid ();
	
	ftruncate (pid_file->fd, 0);
	
	if (write_str != NULL && *write_str != '\0') {
		buf_len = 10 + 1 + strlen (write_str);
		buf = (char*) malloc (buf_len * sizeof (char));
		buf_len = sprintf (buf, "%d %s", pid_inst, write_str);
	}
	else {
		buf = (char*) malloc (10 * sizeof (char));
		buf_len = sprintf (buf, "%d", pid_inst);
	}
	
	if (write (pid_file->fd, buf, buf_len) < 0) {
		free (buf);
		return false;
	}

	pid_file->holder = (pid_t) pid_inst;
	pid_file->rest_str = common_strdup (write_str);
	return true;
}

static bool
common_app_pid_file_read_file (CommonPidFile* pid_file)
{
	int	 pid_inst;
	
	char*			buf;
	int				buf_len;
	char*			buf_rest;

	struct stat		st;
	
	if (fstat (pid_file->fd, &st)) {
		close (pid_file->fd);
		return false;
	}

	buf = (char*) malloc ((st.st_size + 1) * sizeof(char));
	buf_len = read (pid_file->fd, buf, st.st_size);
	
	if (buf_len < 0) {
		free (buf);
		return false;
	}
	
	buf[buf_len] = '\0';

	pid_inst = (int) strtol (buf, &buf_rest, 0);
	if (buf == buf_rest) 
		WRNG ("[LOCK]Lock file does not contains pid.\n");

	pid_file->holder = (pid_t)pid_inst;
	pid_file->rest_str = common_strdup (common_str_skip_space (buf_rest));
	return true;
}

static void
common_app_pid_file_add_cleanup (CommonPidFile* pid_file)
{
	if (pid_file_cleanup == NULL) {
		pid_file_cleanup_alloc = 4;
		pid_file_cleanup = (CommonPidFile**)
							malloc (pid_file_cleanup_alloc * sizeof (void*));

		common_app_at_exit ((ExitFunc)common_app_pid_file_cleanup_exit, NULL, NULL);
	}
	else if (pid_file_cleanup_count == pid_file_cleanup_alloc) {
		pid_file_cleanup_alloc <<= 1;
		pid_file_cleanup = (CommonPidFile**)
							realloc (pid_file_cleanup,
							         pid_file_cleanup_alloc * sizeof (void*));
	}

	pid_file_cleanup[pid_file_cleanup_count++] = pid_file;
}

static void
common_app_pid_file_remove_cleanup (CommonPidFile* pid_file)
{
	int i;
	int j;

	for (i = 0; i < pid_file_cleanup_count; i++ ) {
		if (pid_file_cleanup[i] == pid_file) {
			pid_file_cleanup_count--;
			for (j = i; j < pid_file_cleanup_count; j++)
				pid_file_cleanup[j] = pid_file_cleanup[j + 1];
			break;
		}
	}
}

static void
common_app_pid_file_cleanup_exit	(void)
{
	int i;
	cleanup_mode = true;
	
	for (i = 0; i < pid_file_cleanup_count; i++)
		common_app_pid_file_free (pid_file_cleanup[i]);

	free (pid_file_cleanup);
}


/* function body - option processing - internal *******************************/

bool
common_option_process_void  (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	//doing nothing, but set option_ptr (assumed to be bool) to true.
	* ((bool*)option_ptr) = true;
	return true;
}

bool
common_option_process_bool  (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	bool	value;
	if (opt_arg == NULL) value = true;
	else if (! common_str_is_bool (opt_arg, &value)) return false;

	* ((bool*)option_ptr) = value;
	return true;
}


bool
common_option_process_char  (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	if ((opt_arg[1] != '\0') || (opt_arg[0] == '\0')) return false;
	* ((char*) option_ptr) = opt_arg[0];
	return true;
}

bool
common_option_process_int   (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	int	value;
	
	if (! common_str_is_integer (opt_arg, &value)) return false;
	* ((int*) option_ptr) = value;
	return true;
}


bool
common_option_process_enum  (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	CommonOptionEnumData*   type_data_enum;
	int i;

	if (opt_arg == NULL) return false;
	
	type_data_enum = (CommonOptionEnumData*) type_data;

	for (i = 0; i < type_data_enum->str_list_count; i++) {
		if (strcmp (opt_arg, type_data_enum->str_list[i]) == 0) {
			* ((int*)option_ptr) = i;
			return true;
		}
	}

	return false;
}

bool
common_option_process_str   (void*			option_ptr,
                             const void*	type_data,
                             const char*	opt_arg)
{
	* ((const char**)option_ptr) = opt_arg;
	return true;
}
