/*
 * common_app.h
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler and device-autorotator are free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 * 
 * touchscreen-autodisabler and device-autorotator are distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#ifndef __COMMON_APP_HEADER
#define __COMMON_APP_HEADER

#include <unistd.h>
#include <getopt.h>

/* enum declarations - application core support *******************************/

/*
 * CommonDaemonResult:
 * @COMMON_DAEMON_RESULT_OK:		Daemonization was succesful.
 * @COMMON_DAEMON_RESULT_PARENT:	This is parent process.
 * @COMMON_DAEMON_RESULT_FAIL:		Daemonization failed.
 * 
 * Result of common_app_daemonize()
 */
typedef enum _CommonDaemonResult {
	COMMON_DAEMON_RESULT_OK,
	COMMON_DAEMON_RESULT_PARENT,
	COMMON_DAEMON_RESULT_FAIL
} CommonDaemonResult;

/*
 * CommonType:
 * @COMMON_TYPE_VOID:   void type.
 * @COMMON_TYPE_BOOL:	boolean type.
 * @COMMON_TYPE_CHAR:   char type.
 * @COMMON_TYPE_INT:	integer type.
 * @COMMON_TYPE_ENUM:   enumeration type.
 * @COMMON_TYPE_STRING: string type.
 * 
 * Possible types of option.
 * 
 * For %COMMON_TYPE_VOID, option struct just holds a #bool for it.
 * this type will not have boolean default value, just being checked that
 * it was in option or not.
 * 
 * For %COMMON_TYPE_ENUM, #CommonOptionItem.type_data will point to a
 * #CommonOptionEnumData.
 */
typedef enum _CommonType {
	COMMON_TYPE_VOID,
	COMMON_TYPE_BOOL,
	COMMON_TYPE_CHAR,
	COMMON_TYPE_INT,
	COMMON_TYPE_ENUM,
	COMMON_TYPE_STRING,

	COMMON_TYPE_COUNTS
} CommonType;

/* struct declarations - application core support *****************************/

/*
 * CommonPidFile:
 * @fd:			fild description
 * @file_path:  path of pid file.
 * @holder:		pid that written on pid file. (generally pid of holding process)
 * @rest_str:   rest or string that written on pid file.
 * 
 * Represents pid file.
 * 
 * If pid file is locked by other process, this struct will have different
 * pid at @holder, rather than getpid()
 *
 * // Example. checking pid file is locked or not.
 * CommonPidFile* file = common_app_pid_file (pid_path, NULL);
 * 
 * if (file->holder != getpid ()) {
 *	 printf ("pid file is locked!!\n")
 *	 return;
 * }
 */
typedef struct _CommonPidFile {
	int			fd;
	char*		file_path;
	
	pid_t		holder;
	char*		rest_str;
} CommonPidFile;

/* struct declarations - option processing ************************************/

/*
 * CommonOptionItem:
 * @description:	description of option.
 * @type:			type of option.
 * @type_data:		additional information for type.
 * @default_data:   currently unused.
 * @offset:			offset of member in structure.
 * 
 * A structure for additional information (description, type information...)
 * for option processing.
 * 
 * an result of offsetof() can be used as @offset.
 */
typedef struct _CommonOptionItem {
	const char* description;
	
	CommonType  type;
	void*		type_data;
	void*		default_data;
	
	size_t		offset;
} CommonOptionItem;

/*
 * CommonOptionEnumData:
 * @str_list: (array length=str_list_count):	array of possible option string.
 * @str_list_count:								length of @str_list.
 * 
 * Additional information for %COMMON_TYPE_ENUM.
 */
typedef struct _CommonOptionEnumData {
	const char**	str_list;
	int				str_list_count;
} CommonOptionEnumData;


typedef struct _CommonOpionResultBase {
	char*			command_line;
} CommonOptionResultBase;

/* callback declaration *******************************************************/

/*
 * ExitFunc:
 * @user_data:  user data.
 * 
 * This is prototype for common_app_at_exit().
 */ 
typedef void	(*ExitFunc)		(void*			user_data);

/*
 * SigFunc:
 * @signal: signal.
 * 
 * This is prototype for signal handlers.
 */
typedef void	(*SigFunc)		(int			signal);

/* function declaration - application core support ****************************/

void				common_app_at_exit		(ExitFunc		func,
						                     void*			user_data,
						                     CommonFreeFunc free_func);

bool				common_app_register_sig (SigFunc		func,
				                             ...);

char*				common_app_home_path	(void);

CommonPidFile*		common_app_pid_file		(const char*	lock_file_path,
			                                 const char*	write_str,
				                             bool			replace);

void				common_app_pid_file_free(CommonPidFile* file);

CommonDaemonResult	common_app_daemonize	(pid_t*			child);

void				common_app_replace		(pid_t			instance);

/* function declaration - command line option processing */



char*				common_option_generate_short(const struct option*		option_longopt);

bool				common_option_process		(void*						option_struct,
						                         const struct option*		option_longopt,
						                         const CommonOptionItem*	option_item,
						                         int						argc,
						                         char**						argv);

void				common_option_help			(const struct option*		option_longopt,
							                     const CommonOptionItem*  option_item);
#endif  //COMMON_APP_HEADER