/*
 * common_util.c
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler and device-autorotator are free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 * 
 * touchscreen-autodisabler and device-autorotator are distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include "common_util.h"

/*
 * common_strdup:
 * @str:	string to be copied.
 * 
 * Makes a copy of @str on memory. Free it after use.
 * 
 * Returns: (transfer full): copy of @str.
 */
char*
common_strdup  (const char*	str)
{
	size_t		str_len;
	char*		result;

	if (str == NULL) return NULL;

	str_len = strlen (str);
	result = (char*) malloc ((str_len + 1) * sizeof(char));

	memcpy (result, str, (str_len + 1));
	return result;
}

/*
 * common_str_skip_space:
 * @str:	string to skip space.
 * 
 * Finds first occurance of non-space character on @str.
 * 
 * Returns: (transfer none): first position of non-space character on @str.
 */
const char*
common_str_skip_space (const char*	str)
{
	const char* result;

	for (result = str;
	     isspace(*result);
	     result++) {}

	return result;
}

/*
 * common_str_is_integer:
 * @str:			string to check.
 * @result: (out):  the integer result, if @str is integer.
 *
 * Checks that @str is integer.
 * 
 * Returns: whether @str is integer.
 */
bool
common_str_is_integer (const char*  str,
                	   int*			result)
{
	char* str_pos;
	*result = (int) strtol(str, &str_pos, 10);
	return (*str_pos == '\0');
}

/*
 * common_str_is_bool:
 * @str:			string to check.
 * @result: (out):  the bool result, if @str means logical value.
 *
 * Checks that @str is boolean.
 * It means if @str means following, (case insensitive) it returns true.
 * - y			n
 * - yes		no
 * - t			f
 * - true		false
 * - (number)   0
 *
 * And also returns the value of @str into @result.
 * 
 * Returns: whether @str is boolean.
 */
bool
common_str_is_bool	(const char*	str,
             		 bool*			result)
{
	int		int_value;
	char*   str_l;
	char*   str_d;

	bool	is_bool;
	
	str_l = (char*) malloc (strlen(str) + 1);

	sscanf (str, "%s", str_l);
	str_d = common_str_down (str_l);

	is_bool = true;
	if ((strcmp(str_l, "y") == 0) ||		(strcmp(str_l, "yes") == 0) ||
		(strcmp(str_l, "t") == 0) ||		(strcmp(str_l, "true") == 0))
		*result = true;
	else if ((strcmp(str_l, "n") == 0) ||   (strcmp(str_l, "no") == 0) ||
		(strcmp(str_l, "f") == 0) ||		(strcmp(str_l, "false") == 0))
		*result = false;
	else if (common_str_is_integer (str_l, &int_value))
		*result = (int_value != 0);
	else is_bool = false;

	free (str_d);
	free (str_l);
	return is_bool;
}

/*
 * common_strv_get_index:
 * @strv:   string vector.
 * @str:	string to get index.
 * 
 * Looks up @strv for index of @str in it.
 * 
 * Returns: index of @str in @strv, or -1 if @str is not in @strv.
 */
int
common_strv_get_index (const char**	strv,
			    	   const char*	str)
{
	int i;

	for (i = 0; strv[i] != NULL; i++)
		if (strcmp (strv[i], str) == 0) return i;

	return -1;
}

/*
 * common_str_down:
 * @str:	strint to downcase.
 * 
 * creates downcased string from @str.
 * 
 * Returns: (transfer full): newly created string.
 */
char*
common_str_down (const char*   str)
{
	int		i;
	int		len;

	char*   result;

	len = strlen (str);

	result = malloc ((len + 1) * sizeof (char));

	for (i = 0; i < len; i++)
		result[i] = tolower (str[i]);
	result [len] = '\0';

	return result;
}

/*
 * common_str_join_space:
 * @strv: (array length=strv_count):	string array to put together.
 * @strv_count:							number of strings in @strv
 * 
 * Combines a group of string into a string.
 * 
 * Returns: combined @strv, which separated by ' '
 */
char*
common_str_join (char* const*   strv,
                 int			strv_count,
                 char			join_char)
{
	int		i;
	int		str_length;

	char*   result;
	char*   result_ptr;

	if (strv_count == 0) return NULL;

	str_length = strv_count;
	for (i = 0; i < strv_count; i++) str_length += strlen(strv[i]);

	result = (char*) malloc (str_length * sizeof (char));

	result_ptr = result + sprintf(result, "%s", strv[0]); 
	for (i = 1; i < strv_count; i++)
		result_ptr += sprintf (result_ptr, "%c%s", join_char, strv[i]);

	return result;
}

/* function body - simple closure support *************************************/

/*
 * common_closure_set:
 * @closure:	closure structure.
 * @func:		function to set.
 * @user_data:  user_data to use with @func.
 * @free_func:  function to free @user_data.
 * 
 * sets closure with given function.
 * 
 * @user_data will be attached in invocation. but, @func is not make use of
 * @user_data, it would be just ignored.
 * @free_func will be called when closure gets finalized. If @user_data is from
 * static or const area, or NULL, @free_func also may be NULL.
 * 
 * This function is used in "static allocated fashion". (allocated in static
 * area or struct.)
 * 
 * //Example: struct allocated closure.
 * typedef struct _Asdfqwer {
 *    int result;
 *    CommonClosure closure;
 * }
 *
 * ...
 * 
 * Asdfqwer item;
 * item.result = 0;
 * common_closure_set ( & (item.closure), ..., ..., ...);
 */
void
common_closure_set (CommonClosure*  closure,
					void*			func,
					void*			user_data,
					CommonFreeFunc  free_func)
{
	closure->func = func;
	closure->user_data = user_data;
	closure->free_func = free_func;
}

/*
 * common_closure_fini:
 * @closure:	closure to reset
 * 
 * frees #CommonClosure.user_data and unset @closure.
 * 
 * It can be used finalize @closure or reuse new constructed @closure.
 */
void
common_closure_fini	(CommonClosure* closure)
{
	if ((closure->func != NULL) && (closure->free_func != NULL))
		closure->free_func (closure->user_data);
	common_closure_set (closure, NULL, NULL, NULL);
}

/*
 * common_closure_new:
 * @func:		function to set.
 * @user_data:  user_data to use with @func.
 * @free_func:  function to free @user_data.
 * 
 * constructs closure with given function.
 * 
 * for general information about argments, see common_closure_set()
 *
 * The pointer result of this function should be freed with common_closure_free()
 *
 * This function is used in "construct fashion". (Dynamically construct)
 *
 * //Example: Construct closure arrays.
 * 
 * CommonClosure**  closures;
 * int				closure_count;
 * int				closure_alloc;
 * 
 * ...
 * //When adding closure.
 * 
 * if (closure_count == closure_alloc) {
 *	  closure_alloc *= 2;
 *    closures = (CommonClosure**) realloc (closures, closure_alloc * sizeof (void*));
 * }
 * closures[closure_count++] = common_closure_new (..., ..., ...);
 */
CommonClosure*
common_closure_new (void*			func,
					void*			user_data,
					CommonFreeFunc  free_func)
{
	CommonClosure* result;
	result = (CommonClosure*) malloc (sizeof (CommonClosure));

	common_closure_set (result, func, user_data, free_func);
	return result;
}
/*
 * common_closure_free:
 * @closure:	closure to free.
 * 
 * frees @closure that constructed by common_closure_new().
 * @user_data also be freed.
 */
void
common_closure_free	(CommonClosure* closure) {
	if ((closure->func != NULL) && (closure->free_func != NULL))
		closure->free_func (closure->user_data);
	free (closure);
}
