/*
 * common_util.h
 * Copyright (C) 2014 WSID <jongsome@naver.com>
 * 
 * touchscreen-autodisabler and device-autorotator are free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 * 
 * touchscreen-autodisabler and device-autorotator are distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#ifndef __COMMON_UTIL_HEADER
#define __COMMON_UTIL_HEADER

#include <stdio.h>
#include <stdlib.h>

#include <stdbool.h>

#include "config.h"

/* Debug definitions **********************************************************/

/* DEBUG:
 *
 * Switch macro to enable and disable debug statements.
 * this depends on config.h.
 * Therefore configure.ac need to have following statements, so that debug
 * statement can be turn on or off on configuring time.
 * 
 * dnl Example - autoconf
 * 
 * AC_ARG_ENABLE([debug],
 * 				 [AS_HELP_STRING([--enable-debug],
 * 								 [Enable to print debugging messages])],
 * 				 [AC_DEFINE([DEBUG], [], [Enabling debug macros])])
 */

#ifdef DEBUG
	#define DMSG(...)   fprintf(stderr, "DEBUG: " __VA_ARGS__)
#else
	#define DMSG(...)   /**/
#endif

#define WRNG(...)		fprintf(stderr, "WARNING: " __VA_ARGS__)
#define ERR(code, ...)  do{fprintf(stderr, "ERROR: " __VA_ARGS__); exit(code);}while(false)

#define BSTR(b)			((b) ? "true " : "false")

#define EMSTR(s)		((((s) != NULL) && (*(s) != '\0')) ? (s) : "(empty)")

/* struct declarations - simple callback support ******************************/
typedef void	(*CommonFreeFunc)	   (void*			user_data);

typedef struct _CommonClosure {
	void*			func;
	void*			user_data;
	CommonFreeFunc  free_func;
} CommonClosure;


/* function declarations - utility functions **********************************/

char*		common_strdup				(const char*	str);

const char*	common_str_skip_space		(const char*	str);

bool		common_str_is_integer		(const char*	str,
					                     int*			result);

bool		common_str_is_bool			(const char*	str,
						                 bool*			result);

int			common_strv_get_index		(const char**   strv,
						                 const char*	str);

char*		common_str_down				(const char*	str);

char*		common_str_join				(char* const*   strv,
						                 int			strv_count,
						                 char			join_char);

/* function declarations - simple callback support ****************************/
void		common_closure_set			(CommonClosure* closure,
					                     void*			func,
					                     void*			user_data,
					                     CommonFreeFunc free_func);

void		common_closure_fini			(CommonClosure* closure);

CommonClosure* common_closure_new		(void*			func,
					                     void*			user_data,
					                     CommonFreeFunc free_func);

void		common_closure_free			(CommonClosure* closure);

/*
 * FTYPE:
 * @R:		Return type of callback type.
 * @...:	Argument types of callback type.
 * 
 * Expands function type.
 * 
 * Returns: expands into function pointer type, so that it can be used to cast
 *  function pointers.
 */
#define FTYPE(R,...)		R(*)(__VA_ARGS__, void*)

/*
 * COMMON_CLOSURE_INV:
 * @FT:		Function type.
 * @C:		Pointer to a #CommonClosure structure.
 * @...:	Arguments to be supplied to @C.
 * 
 * Invokes @C as @FT with given arguments. This will attach
 * #CommonClosure.user_data on invocation.
 * You cannot use to invoke as (ReturnType(*)(void*)). For that, use
 * #COMMON_CLOSURE_INV_VOID.
 * 
 * Returns: Invocation result, or nothing if @FT is type of void (*)(...)
 */
#define COMMON_CLOSURE_INV(FT, C,...) \
	if((C)->func != NULL) (((FT)((C)->func)) (__VA_ARGS__, (C)->user_data))

/*
 * COMMON_CLOSURE_INV_VOID:
 * @R:		Return type.
 * @C:		Pointer to a #CommonClosure structure.
 * 
 * Invokes @C as (@R(*)(void*)). This will attach #CommonClosure.user_data on
 * invocation.
 * 
 * Returns: Invocation result (Type of @R), or nothing if @R is "void".
 */
#define COMMON_CLOSURE_INV_VOID(R, C) \
	if((C)->func != NULL) (((R(*)(void*))((C)->func)) ((C)->user_data))

#endif  //COMMON_UTIL_HEADER