
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <inttypes.h>

#include <sys/types.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <X11/extensions/Xrandr.h>

#include <X11/extensions/XInput.h>
#include <X11/extensions/XInput2.h>

#include <libintl.h>

#include "dar_device.h"

#define _(s)	gettext (s)
#define N_(s)   (s)

/* function prototype - internal **********************************************/

void		dar_device_fini			(void);

/* function prototype - device functions **************************************/

RROutput	dar_device_get_output   (const char*	name);

const char* dar_device_output_name  (RROutput		output);

/* internal constants *********************************************************/
#define DEV_ARRAY_INITIAL_ALLOC 4

/* static variables ***********************************************************/

static Display*		display;
static Window		window;

static int			xrr_ev_base;
static int			xrr_err_base;

static int			xi_opcode;
static int			xi_event_code;
static int			xi_error_code;

static Atom			atom_swap;
static Atom			atom_flip;

static XDeviceInfo* dev_info;
static int			dev_info_len;

static XRRScreenResources*		scr_res;

static XRROutputInfo**			out_info;

static RROutput					dev_watch = None;
static XRROutputInfo*			dev_watch_info = NULL;

static XID*			dev_array;
static int			dev_array_len;
static int			dev_array_alloc;

/* function body **************************************************************/

void	
dar_device_init (void)
{
	char*		display_name;

	int			i;

	display_name = getenv ("DISPLAY");

	display = XOpenDisplay (display_name);
	window = DefaultRootWindow (display);

	if (! XRRQueryExtension (display, &xrr_ev_base, &xrr_err_base)) {
		fprintf (stderr, _("[ERR] XRandR is not available.\n"));
		exit (-1);
	}
	
	if (! XQueryExtension (display, "XInputExtension", &xi_opcode, &xi_event_code, &xi_error_code)) {
		fprintf (stderr, _("[ERR] XInput is not available.\n"));
		exit (-1);
	}


	scr_res = XRRGetScreenResources (display, window);

	out_info = (XRROutputInfo**)
		malloc (scr_res->noutput * sizeof (XRROutputInfo*));

	for (i = 0; i < scr_res->noutput; i++)
		out_info[i] = XRRGetOutputInfo (display, scr_res, scr_res->outputs[i]);
	
	dev_info = XListInputDevices (display, &dev_info_len);
	
	atom_swap = XInternAtom (display, "Evdev Axes Swap", false);
	atom_flip = XInternAtom (display, "Evdev Axis Inversion", false);

	dev_array = (XID*) malloc (DEV_ARRAY_INITIAL_ALLOC * sizeof(XID));
	dev_array_len = 0;
	dev_array_alloc = DEV_ARRAY_INITIAL_ALLOC;

	atexit (dar_device_fini);
}


void
dar_device_set_watch (const char*		watched)
{
	int i;

	dev_watch = None;
	dev_watch_info = NULL;
	for (i = 0; i < scr_res->noutput; i++) {
		if (strcmp(watched, out_info[i]->name) == 0) {
			dev_watch_info = out_info[i];
			dev_watch = scr_res->outputs[i];
			break;
		}
	}
	if (dev_watch == None) {
		fprintf (stderr, _("[ERR] Output \"%s\" not found.\n"), watched);
		exit (-1);
	}
}

void
dar_device_add_target (const char*		target)
{
	bool				device_is_xid;
	
	const XDeviceInfo*	device_info;
	XID					device_id;
	
	const XDeviceInfo* devinfo_pos;
	const XDeviceInfo* devinfo_end;

	int					i;
	int					i_len;

	device_is_xid = true;
	i_len = strlen (target);
	for (i = 0; i < i_len; i++) {
		if (! isdigit(target[i])) {
			device_is_xid = false;
			break;
		}
	}

	if (device_is_xid)
	     device_id = (XID) atoi (target);

	
	{
		device_info = NULL;
		devinfo_end = dev_info + dev_info_len;
		for (devinfo_pos = dev_info;
			 devinfo_pos < devinfo_end;
			 devinfo_pos ++) {
			if (((!device_is_xid) && (strcmp(target, devinfo_pos->name) == 0)) ||
			    ((device_is_xid) && (device_id == devinfo_pos->id))) {
				device_info = devinfo_pos;
				break;
			}
		}
		if (device_info == NULL) {
			fprintf (stderr, _("[ERR] Device \"%s\" not found.\n"), target);
			exit (-1);
		}
		
		device_id = device_info -> id;
	}

	if (dev_array_len + 1 == dev_array_alloc) {
		dev_array_alloc *= 2;
		dev_array = (XID*) realloc (dev_array, dev_array_alloc * sizeof (XID));
	}
	
	dev_array[dev_array_len] = device_id;
	dev_array_len ++;
}

void
dar_device_watch (DarDeviceCallback  callback,
		          void*				user_data)
{
	XEvent						event;
	XRRNotifyEvent*				ev_notify;
	XRRCrtcChangeNotifyEvent* ev_outchange_notify; 

	long						event_mask;
	Rotation					rotation_x;
	DarRotation					rotation_dar;

	int							fd_display;
	fd_set						fd_display_set;

    event_mask = RRCrtcChangeNotifyMask;

    XRRSelectInput (display, window, event_mask);

	fd_display = ConnectionNumber (display);

	while (true) {
		FD_ZERO (&fd_display_set);
		FD_SET (fd_display, &fd_display_set);

		if ((! XPending (display)) &&
		    (select (fd_display+1, &fd_display_set, NULL, NULL, NULL)))
			continue;

		while (XPending (display)) {
			XNextEvent (display, &event);

			if (event.type == xrr_ev_base + RRNotify) {
				ev_notify = (XRRNotifyEvent*) &event;

				if (ev_notify->subtype == RRNotify_CrtcChange) {
					ev_outchange_notify = (XRRCrtcChangeNotifyEvent*) ev_notify;

					fprintf (stderr, "Crtc:%d\n", 
					         (int)(ev_outchange_notify->crtc));
					
					if (ev_outchange_notify->crtc != dev_watch_info->crtc) {
						fprintf (stderr, "event discarded\n");
						break;
					}
				
					// I wrote monsterous ev transform codes.
					rotation_x = ev_outchange_notify->rotation;

					rotation_dar =
						(rotation_x == RR_Rotate_0)   ? DAR_ROTATION_0 :
						(rotation_x == RR_Rotate_90)  ? DAR_ROTATION_90 :
						(rotation_x == RR_Rotate_180) ? DAR_ROTATION_180 :
						(rotation_x == RR_Rotate_270) ? DAR_ROTATION_270 : -1;

					if (rotation_dar != -1)
						callback (rotation_dar, user_data);
				}
			}
		}
	}
}

void
dar_device_rotate_target (DarRotation rotation) {
	uint8_t		flip[2];
	uint8_t		swap;
	
	int			i;

	flip[0] = dar_x_flip [rotation];
	flip[1] = dar_y_flip [rotation];

	swap = dar_swap [rotation];

	for (i = 0; i < dev_array_len; i++) {
		XIChangeProperty (display,
		                  dev_array[i],
		                  atom_flip,
		                  XA_INTEGER,
		                  8,
		                  PropModeReplace,
		                  (unsigned char*)flip,
		                  2);

		
		XIChangeProperty (display,
		                  dev_array[i],
		                  atom_swap,
		                  XA_INTEGER,
		                  8,
		                  PropModeReplace,
		                  (unsigned char*)&swap,
		                  1);
	}

	XSync (display, false);
}


/* internal functions *********************************************************/

void
dar_device_fini	(void)
{
	int i;

	for (i = 0; i < scr_res->noutput; i++)
		XRRFreeOutputInfo (out_info[i]);

	XRRFreeScreenResources (scr_res);

	free (out_info);
	free (dev_array);
	XFreeDeviceList (dev_info);
	XCloseDisplay (display);
}

/* device functions ***********************************************************/

RROutput
dar_device_get_output (const char*	name)
{
	int i;
	for (i = 0; i < scr_res->noutput; i++)
		if (strcmp (name, out_info[i]->name) == 0) return scr_res->outputs[i];
	return None;
}

const char*
dar_device_output_name (RROutput	output)
{
	int i;
	for (i = 0; i < scr_res->noutput; i++)
		if (scr_res->outputs[i] == output) return out_info[i]->name;
	return NULL;
}