/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * main.c
 * Copyright (C) 2014 WSID <wsid@wsid-pc>
 * 
 * device-autorotator is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * device-autorotator is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdbool.h>
#include <string.h>

#include <getopt.h>
#include <pwd.h>

#include <signal.h>

#include <libintl.h>

#include "common_util.h"
#include "common_app.h"

#include "dar_def.h"
#include "dar_device.h"


#define _(s)	gettext(s)
#define N_(s)   (s)

typedef enum _DarOptionID {
	DAR_OPTION_HELP = 'h',
	DAR_OPTION_VERSION = 'v',
	
	DAR_OPTION_DAEMONIZE = 'd',
	DAR_OPTION_REPLACE = 'r'
} DarOptionID;

typedef struct _DarOption {
	CommonOptionResultBase		base;
	bool	help;
	bool	version;

	bool	daemonize;
	bool	replace;
} DarOption;

/* function declarations ******************************************************/
void	dar_init (const DarOption* option);

void	dar_interrupt (void);

void	dar_help (void);
void	dar_version (void);

void	dar_daemonize (void);

void	dar_lock (const char* option, bool replace);

/* constant declarations ******************************************************/

static const struct option   option_long[] = {
	{"help",		no_argument,		NULL, DAR_OPTION_HELP},
	{"version",		no_argument,		NULL, DAR_OPTION_VERSION},
	{"daemonize",   optional_argument,  NULL, DAR_OPTION_DAEMONIZE},
	{"replace",		optional_argument,  NULL, DAR_OPTION_REPLACE},

	{NULL, 0, NULL, 0}
};

static const CommonOptionItem   option_item[] = {
	{N_("Print this help message"),
		COMMON_TYPE_VOID,   NULL,	NULL,		offsetof(DarOption, help) },
	{N_("Print version information"),
		COMMON_TYPE_VOID,   NULL,	NULL,		offsetof(DarOption, version) },
	{N_("Run in background"),
		COMMON_TYPE_BOOL,   NULL,	NULL,		offsetof(DarOption, daemonize) },
	{N_("Replace current running instance"),
		COMMON_TYPE_BOOL,   NULL,	NULL,		offsetof(DarOption, replace) },
};


/* function body **************************************************************/

/*
 * main:
 * @argc: count of commandline arguments.
 * @argv: content of commandline arguments.
 * 
 * application entry point.
 * 
 * Returns: 0 on success. Otherwise, exit code.
 */
int
main (int argc, char *argv[])
{
	DarOption   option;
	int i;

	setlocale(LC_ALL, "");
	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain (GETTEXT_PACKAGE);

	memset (&option, 0, sizeof(DarOption));
	common_option_process ((CommonOptionResultBase*)&option,
	                       option_long,
	                       option_item,
	                       argc, argv);

	dar_init (&option);

	if (argc - optind < 1) {
		fprintf (stderr, _("Need output name.\n"));
		return -1;
	}
	else if (argc - optind < 2) {
		fprintf (stderr, _("need device name\n"));
		return -1;
	}
	
	dar_device_set_watch (argv[optind]);
	
	for (i = optind + 1; i < argc; i++)
		dar_device_add_target (argv[i]);

	dar_device_watch ((DarDeviceCallback)dar_device_rotate_target, NULL);
	
	return 0;
}

/*
 * dar_init:
 * @option_str: options combined into one string.
 * @option:		option structure.
 * 
 * initialize application.
 */
void
dar_init (const DarOption* option)
{
	if (option->help) dar_help ();
	if (option->version) dar_version ();

	if (option->daemonize) dar_daemonize ();
	
	if (! common_app_register_sig((SigFunc)dar_interrupt, SIGINT, SIGTERM, 0)) {
		fprintf (stderr, _("[ERR] Signal handler install failed.\n"));
		exit (-1);
	}

	dar_lock (option->base.command_line, option->replace);
	
	dar_device_init ();
}

/*
 * dar_interrupt:
 * 
 * Exits program with succesful exit.
 */
void
dar_interrupt (void)
{
	exit (0);
}



/*
 * dar_help:
 * 
 * Prints help message and exit.
 */
void
dar_help (void)
{
	printf (_("device_autorotator:\n"
	        "  autorotate devices by screen rotation.\n"
	        "\n"
	        "device_autorotator \"output\" \"devices\" ...\n"
	        "output: output to be watched.\n"
	        "devices: a list of devices to be associated with output\n"
	        "\n"
	        "name of output can be obtained by xrandr\n"
	        "name of devices can be obtained by xinput\n"
	        "\n"));
	common_option_help (option_long, option_item);
	exit (0);
}

/*
 * dar_version:
 * 
 * Prints version information and exit.
 */
void
dar_version (void)
{
	printf (_("device_autorotator 0.1.1 - 2014, Copyrights WSID <jongsome@naver.com>\n"));
	exit (0);
}



/*
 * dar_daemonize:
 * 
 * Daemonize this application.
 * 
 * This slightly wraps common_app_daemonize ()
 */
void
dar_daemonize (void)
{
	CommonDaemonResult  result;
	pid_t				child;

	result = common_app_daemonize (&child);

	switch (result) {
		case COMMON_DAEMON_RESULT_OK:
			break;
			
		case COMMON_DAEMON_RESULT_PARENT:
			printf (_("New device_autorotator spawned as daemon.\n"
			          "Pid: %d\n"),
			        (int)child);
			exit (0);
			
		case COMMON_DAEMON_RESULT_FAIL:
			perror ("device_autorotator");
			fprintf (stderr, _("[DAEMON] Failed to daemonize."));
			exit (-1);
	}
}



/*
 * dar_lock:
 * @option:	 options combined into one string.
 * @replace: whether to terminate running instance if pid file is locked.
 * 
 * Create/Locks pid file for application.
 */
void
dar_lock (const char* option, bool  replace)
{
	CommonPidFile*  pid_file;

	char*			pid_path_part[2];
	char*			pid_path;

	pid_path_part[0] = common_app_home_path ();
	pid_path_part[1] = ".device_autorotator.pid";
	
	pid_path = common_str_join (pid_path_part, 2, '/');
	pid_file = common_app_pid_file (pid_path, option, replace);
	free (pid_path);
	
	if (pid_file == NULL) {
		fprintf (stderr, _("[LOCK] Cannot lock!\n"));
		perror ("device_autorotator");
		exit (-1);
	}

	if (pid_file->holder != getpid()) {
		printf (_("device_autorotator instance is running.\n"
		        "pid: %5d, option: %s\n"),
		        (int)pid_file->holder, pid_file->rest_str);

		common_app_pid_file_free (pid_file);
		exit (0);
	}
	common_app_at_exit ((ExitFunc)common_app_pid_file_free, pid_file, NULL);
}